﻿using System;
using System.Linq;
using System.Linq.Expressions;
using System.Collections.Generic;
using System.Reflection;
using NHibernate;

namespace NMA.Infrastructure.NHibernate.Repository
{
    using NMA.Domain.Model;
    using NMA.Domain.Model.Repository;
    using NMA.Infrastructure.Logging;
    using NMA.Infrastructure.NHibernate.Container;
    using NMA.Infrastructure.NHibernate.UnitOfWork;
    using NMA.Infrastructure.DBC;
    using NMA.Infrastructure.NHibernate.DomainObject;
    using NMA.Domain.Shared.Paging;

    public class AuditRepository : BaseRepository<IAudit>, IAuditRepository
    {
        #region variables        

        protected IUnitOfWorkFactory UnitOfWorkFactory { get; private set; }

        private static readonly ILog _logger = LogManager.GetLogger(MethodBase.GetCurrentMethod().DeclaringType);

        #endregion

        #region contructors        

        public AuditRepository() : this(IoC.GetInstance<IUnitOfWorkFactory>())
        { }

        internal AuditRepository(IUnitOfWorkFactory unitOfWorkFactory) 
        {
            UnitOfWorkFactory = unitOfWorkFactory;
        }

        #endregion

        #region implement functions

        public IEnumerable<IAudit> All()
        {
            using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
            {
                ICriteria critreria = uow.Session.CreateCriteria<Audit>();
                uow.Commit();
                return critreria.List<Audit>();
            }
        }

        public IAudit GetById(Guid id)
        {
            Check.Assert(id != null, "Id is null");

            try
            {
                IAudit audit;

                using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
                {
                    audit = uow.Session.Get<Audit>(id);
                    uow.Commit();
                }

                return audit;
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                throw ex;
            } 
        }     

        public Guid Add(IAudit entity)
        {
            Check.Assert(entity != null, "Audit's object is null");

            try
            {
                object result = null;

                using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
                {
                    result = uow.Session.Save(entity);

                    uow.Session.Flush();
                    uow.Session.Clear();

                    uow.Commit();
                }

                return (Guid)result;
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                throw ex;
            }
        }

        public void Remove(IAudit entity)
        {
            Check.Assert(entity != null, "Audit's object is null");

            try
            {
                using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
                {
                    uow.Session.Delete(entity);
                    uow.Commit();
                }
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                throw ex;
            }
        }

        public void Remove(Guid id)
        {
            Check.Assert(id != null, "Audit's id is null");

            try
            {
                var entity = GetById(id);

                if (entity == null)
                    throw new Exception("Audit's entity is null");

                using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
                {
                    uow.Session.Delete(entity);
                    uow.Commit();
                }
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                throw ex;
            }
        }

        public void Update(IAudit entity)
        {
            Check.Assert(entity != null, "Audit's entity is null");

            try
            {
                using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
                {
                    uow.Session.Update(entity);
                    uow.Commit();
                }
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                throw ex;
            } 
        }

        public IEnumerable<IAudit> GetBy(Expression<Func<IAudit, bool>> condition)
        {
            Check.Assert(condition != null, "Condition for searching audit is null");

            try
            {
                IEnumerable<IAudit> result = null;
                IList<IAudit> listOfAudit = null;

                using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
                {
                    // creating criteria for this condition
                    ICriteria criteria = uow.Session.CreateCriteria(typeof(IAudit));
                    // getting result from criteria
                    listOfAudit = criteria.List<IAudit>();

                    if (listOfAudit == null)
                        throw new Exception("List of audit is null");

                    result = listOfAudit.Where<IAudit>(condition.Compile());
                }

                return result;
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                throw ex;
            } 
        }

        public IPagedList<IAudit> All(int pageIndex, int pageSize)
        {
            using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
            {
                var criteria = uow.Session.CreateCriteria(typeof(Audit));

                return PagedList(criteria, pageIndex, pageSize);
            }
        }
    
        #endregion
    }
}
