﻿using System;
using System.Linq;
using NHibernate;
using NHibernate.Criterion;
using NHibernate.Context;
using System.Linq.Expressions;
using System.Collections.Generic;
using System.Reflection;

namespace NMA.Infrastructure.NHibernate.Repository
{
    using NMA.Domain.Model;
    using NMA.Infrastructure.UnitOfWork;
    using NMA.Infrastructure.NHibernate.UnitOfWork;
    using NMA.Infrastructure.NHibernate.Container;
    using NMA.Infrastructure.NHibernate.DomainObject;
    using NMA.Infrastructure.DBC;
    using NMA.Infrastructure.Logging;
    using NMA.Domain.Model.Repository;
    using NMA.Domain.Shared.Paging;

    public class UserRepository : BaseRepository<IUser>, IUserRepository
    {
        #region variables        

        protected IUnitOfWorkFactory UnitOfWorkFactory { get; private set; }

        private static readonly ILog _logger = LogManager.GetLogger(MethodBase.GetCurrentMethod().DeclaringType);

        #endregion

        #region contructors        

        public UserRepository() : this(IoC.GetInstance<IUnitOfWorkFactory>())
        { }

        internal UserRepository(IUnitOfWorkFactory unitOfWorkFactory) 
        {
            UnitOfWorkFactory = unitOfWorkFactory;
        }

        #endregion

        #region implement functions 

        public IEnumerable<IUser> All()
        {
            using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
            {
                ICriteria critreria = uow.Session.CreateCriteria<User>();
                uow.Commit();
                return critreria.List<User>();
            }
        }

        public IUser GetById(Guid id)
        {
            Check.Assert(id != null, "Id is null");

            try
            {
                IUser user;

                using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
                {
                    user = uow.Session.Get<User>(id);
                    uow.Commit();
                }

                return user;
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                throw ex;
            }
        }

        public Guid Add(IUser entity)
        {
            Check.Assert(entity != null, "User's object is null");

            try
            {
                object result = null;

                using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
                {
                    result = uow.Session.Save(entity);

                    uow.Session.Flush();
                    uow.Session.Clear();

                    uow.Commit();
                }

                return (Guid)result;
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                throw ex;
            } 
        }

        public void Remove(IUser entity)
        {
            Check.Assert(entity != null, "User's object is null");

            try
            {
                using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
                {
                    uow.Session.Delete(entity);
                    uow.Commit();
                }
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                throw ex;
            }
        }

        public void Update(IUser entity)
        {
            Check.Assert(entity != null, "User's entity is null");

            try
            {
                using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
                {
                    uow.Session.Update(entity);
                    uow.Commit();
                }
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                throw ex;
            }
        }

        public IEnumerable<IUser> GetBy(Expression<Func<IUser, bool>> condition)
        {
            Check.Assert(condition != null, "Condition for searching user is null");

            try
            {
                IEnumerable<IUser> result = null;
                IList<IUser> listOfUser = null;

                using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
                {
                    // creating criteria for this condition
                    ICriteria criteria = uow.Session.CreateCriteria(typeof(IUser));
                    // getting result from criteria
                    listOfUser = criteria.List<IUser>();

                    if (listOfUser == null)
                        throw new Exception("List of user is null");

                    result = listOfUser.Where<IUser>(condition.Compile());
                }

                return result;
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message);
                throw ex;
            }
        }

        public IPagedList<IUser> All(int pageIndex, int pageSize)
        {
            using (var uow = UnitOfWorkFactory.GetNewUnitOfWork())
            {
                var criteria = uow.Session.CreateCriteria(typeof(User));

                return PagedList(criteria, pageIndex, pageSize);
            }
        }
   
        #endregion
    }
}